#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "simplecom.h"
#include "dslconf.h"

#define UINT32 uint32_t
#define INT32 int32_t

#define BUFLEN	5000000

#if defined(DEBUG)
#define dprint(...) { \
	fprintf(stderr, "simplecom: %s(): ", __func__); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n"); \
}
#else
#define dprint(...)	/* nothing */
#endif

struct action_map_s {
	const char *name;
	int action;
};

struct action_map_s action_map[] = {
	{ .name = "nop", .action = REQ_NOP },
	{ .name = "call", .action = REQ_EXEC },
	{ .name = "get", .action = REQ_GET_ENV },
	{ .name = "set", .action = REQ_SET_ENV },
	{ .name = "flash", .action = REQ_FLASH_CFG },
	{ .name = "reload", .action = REQ_REEXEC },
	{ .name = "gw", .action = REQ_GET_GW },
	{ .name = "ip", .action = REQ_GET_IP },
	{ .name = "setip", .action = REQ_SET_IP },
	{ .name = "start_logic", .action = REQ_START_LOGIC },
	{ .name = "reload_logic", .action = REQ_RELOAD_LOGIC },
	{ .name = "stop_logic", .action = REQ_STOP_LOGIC },
	{ .name = "fwflash", .action = REQ_FLASH_FW },
	{ .name = "pub_ip", .action = REQ_GET_PUB_IP },
	{ .name = "sync_stat", .action = REQ_SYNC_STAT },
	{ .name = "rate_info", .action = REQ_RATE_INFO },
	{ .name = "ppp_info", .action = REQ_PPP_INFO },
	{ .name = "version", .action = REQ_GET_VERSION },
	{ .name = "fsstamp", .action = REQ_FS_STAMP },
	{ .name = "pppd_log", .action = REQ_GET_PPPD_LOG },
	{ .name = "vdsl_cfg", .action = REQ_SET_VDSL_CFG },
	{ .name = "builtin_vdsl_cfg", .action = REQ_USE_BUILTIN_VDSL_CFG },
	{ .name = "wan_mgr", .action = REQ_WAN_MGR },
	{ .name = "vdsl_cfg_sha1", .action = REQ_VDSL_CFG_SHA1 },
	{ .name = "vdsl_mon_stats", .action = REQ_VDSL_MON_STATS },
	{ .name = "logread", .action = REQ_NOTIFY },
	{ .name = "scmver", .action = REQ_SCMVER },
	{ .name = NULL, .action = -1 },
};

static void print_actions(void)
{
	struct action_map_s *as;

	for (as = action_map; as->name; as++) {
		printf("%s\n", as->name);
	}
}

/* concatenate all strings in argv into a newly allocated char* the caller has to free */
char *implode(int argc, char **argv)
{
	int i, len = 0;
	char *retval;

	if (!argc)
		return NULL;

	/* sum each part plus the appended space (or \0 in the last case) */
	for (i = 0; i < argc; i++)
		len += strlen(argv[i]) + 1;

	retval = malloc(len);
	memset(retval, 0, len);

	strcat(retval, argv[0]);
	for (i = 1; i < argc; i++) {
		strcat(retval, " ");
		strcat(retval, argv[i]);
	}

	return retval;
}

/* XXX: copied from server.c and modified a bit, maybe share code? */
/* Read contents of file pointed to by fp into newly allocated buffer,
 * returning a pointer to it. Caller has to free returned memory. */
char *read_fp_to_buf(FILE *fp, size_t *outbuflen)
{
	char *buf = NULL;
	int buflen = 0, offset = 0, rlen;

	while (fp && !feof(fp)) {
		if (offset >= buflen) {
			buflen += 8192;
			buf = realloc(buf, buflen);
			if (!buf)
				return buf;
		}
		rlen = fread(buf + offset, 1, buflen - offset - 1, fp);
		offset += rlen;
		buf[offset] = 0;
	}

	if (outbuflen)
		(*outbuflen) = offset + 1;
	return buf;
}

float ntohf(float val)
{
	union {
		float f;
		__u32 d;
	} tmp = { .f = val };
	tmp.d = ntohl(tmp.d);
	return tmp.f;
}

void print_vdsl_mon_stats(const void *data, __u32 datalen)
{
	const struct {
	    char LineTypeDesc[40];
	    UINT32 LineType, UsLineRate, DsLineRate,
		   DsAttainableRate, UsAttainableRate;
	    short txPwr, txPwr_FE;

	    INT32  SnrMargin;
	    UINT32 Attn;
	    INT32  SnrAvgMargin, SnrAvg;

	    float  FeAttn[5], FeSNRM[6];
	    unsigned char FeAttnStatus, FeSNRMStatus;

	    UINT32 NeHEC_F, NeHEC_S, FeHEC_F, FeHEC_S,
		   NeCRC_F, NeCRC_S, FeCRC_F, FeCRC_S;
	} *__attribute__((packed)) stats_packed;
	int i;

	if (datalen < sizeof(*stats_packed)) {
		printf("too little data returned?!\n");
		return;
	}

	stats_packed = data;

#define PRINTL(f) printf(#f ": %d\n", ntohl(stats_packed->f))
#define PRINTS(f) printf(#f ": %d\n", ntohs(stats_packed->f))
	printf("LineTypeDesc: %s\n", stats_packed->LineTypeDesc);
	PRINTL(LineType);
	PRINTL(UsLineRate);
	PRINTL(DsLineRate);
	PRINTL(DsAttainableRate);
	PRINTL(UsAttainableRate);
	PRINTS(txPwr);
	PRINTS(txPwr_FE);
	PRINTL(SnrMargin);
	PRINTL(Attn);
	PRINTL(SnrAvgMargin);
	PRINTL(SnrAvg);
	for (i = 0; i < 5; i++)
		printf("FeAttn[%d]: %f\n", i, ntohf(stats_packed->FeAttn[i]));
	for (i = 0; i < 6; i++)
		printf("FeSNRM[%d]: %f\n", i, ntohf(stats_packed->FeSNRM[i]));
	printf("FeAttnStatus: %d\n", stats_packed->FeAttnStatus);
	printf("FeSNRMStatus: %d\n", stats_packed->FeSNRMStatus);
	PRINTL(NeHEC_F);
	PRINTL(NeHEC_S);
	PRINTL(FeHEC_F);
	PRINTL(FeHEC_S);
	PRINTL(NeCRC_F);
	PRINTL(NeCRC_S);
	PRINTL(FeCRC_F);
	PRINTL(FeCRC_S);
}

int run_dslconf_cmd(const struct interface iface, const char *cmd, const char *args)
{
	struct action_map_s *amp;
	struct dslconf_pkt *pkt, *reply;
	int rc;
	FILE *fp;
	char *buf;
	size_t buflen;

	if (!cmd)
		return 1;

	for (amp = action_map; amp->name; amp++) {
		if (!strcmp(cmd, amp->name))
			break;
	}
	if (!amp->name) {
		printf("unknown action %s\n", cmd);
		return 1;
	}

	switch (amp->action) {
	case REQ_FLASH_CFG:
	case REQ_SET_VDSL_CFG:
		if (!args) {
			printf("%s needs a filename as argument\n", amp->name);
			return 1;
		}

		if ((fp = fopen(args, "r")) == NULL) {
			printf("can't open '%s': %s\n", args, strerror(errno));
			return 1;
		}
		buf = read_fp_to_buf(fp, &buflen);
		fclose(fp);

		/* buf contains trailing zero byte, do not send it out */
		pkt = alloc_dslconf_pkt(amp->action, FLAG_REQ, buf, buflen - 1);

		break;
	case REQ_NOTIFY:
		/* go into endless listen mode printing log messages */
		while (1) {
			rc = recv_dslconf_packet(iface, (void **)&buf, &buflen);
			if (rc <= 0)
				continue;
			pkt = (struct dslconf_pkt *)buf;
			switch (pkt->hdr.type) {
			case REQ_NOTIFY:
				printf("NOTIFY: %s\n", pkt->data);
				break;
			case REQ_WARNING:
				printf("WARNING: %s\n", pkt->data);
				break;
			case REQ_ERROR:
				printf("ERROR: %s\n", pkt->data);
				break;
			default:
				/* empty */
				break;
			}
			free(pkt);
		}
		break;
	default:
		pkt = alloc_dslconf_pkt(amp->action, FLAG_REQ, args,
		                        (args ? strlen(args) : 0));
		break;
	}

	if (!pkt) {
		printf("out of memory!\n");
		return 1;
	}

	dprint("sending request of type %d, flags 0x%x, data '%s'\n",
			pkt->hdr.type, pkt->hdr.flags, pkt->data);

	rc = do_transaction(iface, pkt, &reply);
	free(pkt);
	if (rc == -1)
		return rc; /* internal failure */
	if (rc == 1) {
		dprint("Timeout\n");
		return rc;
	}
	if (reply->hdr.flags & FLAG_ACK) {
		dprint("Command Returned OK: ");
		rc = 0;
	} else if (reply->hdr.flags & FLAG_NACK) {
		dprint("Command Returned not OK: ");
		rc = 2;
	}
	if (reply->hdr.type == REQ_VDSL_MON_STATS) {
		print_vdsl_mon_stats(reply->data, reply->hdr.dlen);
	} else if (ntohl(reply->hdr.dlen)) {
		printf("%s\n", reply->data);
	}

	free(reply);
	return 0;
}


/** Read one command from user input and trigger the appropriate transaction.
 * @param[in] iface Interface to send the transaction out on.
 * @return Zero on success, <0 in error (abort) case, >0 on timeout */
int client_loop(struct interface iface)
{
	char input[BUFLEN], *args;

	printf("> ");
	fflush(stdout);
	fgets(input, BUFLEN, stdin);
	if (strlen(input) < 2)		/* never forget the mandatory trailing newline */
		return -1;

	input[strlen(input) - 1] = 0; /* drop trailing newline */
	if (!strcmp(input, "exit"))
		return -2;

	if (!strcmp(input, "help")) {
		printf("Actions:\n");
		print_actions();
		printf("shell - simulate remote shell\n");
		return 0;
	}

	if (!strcmp(input, "shell")) {
		printf("Shell mode, type \"exit\" to return to command mode\n");
		do {
			printf("sh> ");
			fflush(stdout);
			fgets(input, BUFLEN, stdin);
			if (strlen(input) < 2)
				continue;
			input[strlen(input) - 1] = 0;
			if (!strcmp(input, "exit"))
				break;

			run_dslconf_cmd(iface, "call", input);
		} while (1);
	}

	if ((args = strchr(input, ' '))) {
		(*(args++)) = '\0';
	}
	run_dslconf_cmd(iface, input, args);

	return 0;
}

int main(int argc, char **argv)
{
	int rc = 0;
	struct interface *iface;

	if (argc < 2 || !strcmp(argv[1], "--help")) {
		printf("Usage:\n");
		printf("\t%s iface\t\t- interactive mode\n", argv[0]);
		printf("\t%s iface <command> [\"<args>\"]\t- batch use\n", argv[0]);
		printf("\t%s --version\t\t- print version and exit\n", argv[0]);
		printf("\t%s --help\t\t- print this text and exit\n", argv[0]);
		printf("Actions:\n");
		print_actions();
		return 1;
	}
	if (!strcmp(argv[1], "--version")) {
		puts(VERSION);
		return 0;
	}

	if ((iface = init_interface(argv[1], SCPROTO_DSLCONF)) == NULL) {
		printf("initialisation failed\n");
		return 1;
	}

	if (argc < 3) {
		while(1) {
			if ((rc = client_loop(*iface)) == -2)
				break; /* user requested exit */
			if (rc < 0)
				continue;
			if (rc)
				printf("TIMEOUT\n");
		}
	} else {
		char *args = implode(argc - 3, argv + 3);
		rc = run_dslconf_cmd(*iface, argv[2], args);
		free(args);
	}
	exit_interface(iface);
	return rc;
}
