#define __USE_POSIX199506
#define _GNU_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "simplecom.h"
#include "dslconf.h"
#include "server.h"
#include "stealip.h"

struct dslconf_cmd_s *dslconf_cmd[] = {
	&CMD_NOP,
	&CMD_EXEC,
	&CMD_GET_ENV,
	&CMD_SET_ENV,
#ifdef TARGET_AR7
	&CMD_FLASH_CFG,
	&CMD_START_LOGIC,
	&CMD_RELOAD_LOGIC,
	&CMD_STOP_LOGIC,
	&CMD_FLASH_FW,
#endif
	&CMD_REEXEC,
	&CMD_GET_GW,
	&CMD_GET_IP,
	&CMD_SET_IP,
	&CMD_GET_PUB_IP,
	&CMD_SYNC_STAT,
	&CMD_RATE_INFO,
	&CMD_PPP_INFO,
	&CMD_GET_VERSION,
	&CMD_FS_STAMP,
	&CMD_SCMVER,
	&CMD_GET_PPPD_LOG,
#ifdef TARGET_FUSIV
	&CMD_SET_VDSL_CFG,
	&CMD_USE_BUILTIN_VDSL_CFG,
	&CMD_WAN_MGR,
	&CMD_VDSL_CFG_SHA1,
	&CMD_VDSL_MON_STATS,
#endif
};

/** Fork the calling process into background
 * and reopen stdin, stdout and stderr to /dev/null. */
int daemonise(void)
{
	int pid;
	unsigned i;
	int close_fileno[] = {STDIN_FILENO, STDOUT_FILENO, STDERR_FILENO};

	if ((pid = fork()) == -1)
		return 1;

	if (pid) { /* parent */
		signal(SIGCHLD, SIG_IGN);
		_exit(0);
	}

	setsid();

	if ((pid = open("/dev/null", O_RDWR)) == -1) {
		perror("open()");
		_exit(1);
	}
	for (i = 0; i < ARRAY_SIZE(close_fileno); i++) {
		close(close_fileno[i]);
		dup2(pid, close_fileno[i]);
	}
	return 0;
}

static void close_fds_above(unsigned int base)
{
	while (++base < 128)
		close(base);
}

static void reset_signal_handlers(void)
{
	for (int i = 1; i < 32; i++)
		signal(i, SIG_DFL);
}

struct thread_ctx {
	struct interface *iface;
	sem_t datalock;
};

static void *dslconfd_thread(void *arg)
{
	struct thread_ctx *ctx = arg;
	struct listener_callback *lcb;
	int i, cmd_count;

	/* dynamically allocate a listener_callback array
	 * for dslconf listener() routine with terminating zero entry */
	cmd_count = ARRAY_SIZE(dslconf_cmd);
	dnotify("allocating for %d commands\n", cmd_count);
	lcb = calloc(cmd_count + 1, sizeof(struct listener_callback));
	if (!lcb) {
		derror("out of core!");
		return NULL;
	}
	for (i = 0; i < cmd_count; i++)
		lcb[i] = dslconf_cmd[i]->cb;
	lcb[cmd_count].callback = NULL;

	listener(*ctx->iface, lcb, &ctx->datalock);

	/* not reached */
	return NULL;
}

static void *notifier_thread(void *arg)
{
	struct thread_ctx *ctx = arg;
	unsigned i, updval;

	while (1) {
		for (i = 0; i < ARRAY_SIZE(dslconf_cmd); i++) {
			if (!dslconf_cmd[i]->update)
				continue;
			sem_wait(&ctx->datalock);
			dnotify("calling update routine of command %d", dslconf_cmd[i]->cb.type);
			updval = (*dslconf_cmd[i]->update)();
			sem_post(&ctx->datalock);
			if (updval <= 0)
				continue;
			if (!dslconf_cmd[i]->notify)
				continue;
			(*dslconf_cmd[i]->notify)(*ctx->iface);
		}
		sleep(1);
	}

	/* not reached */
	return NULL;
}

int main(int argc, char **argv)
{
	struct thread_ctx ctx;
	pthread_t notifier, dslconfd;
	int loglevel = REQ_WARNING;

	if (argc != 2) {
		printf("Usage: %s iface\n", argv[0]);
		return 1;
	}

	if (!getenv("DEBUG")) {
		if (daemonise())
			return 1;
	} else {
		loglevel = REQ_NOTIFY;
	}

	close_fds_above(STDERR_FILENO);
	reset_signal_handlers();

	if ((ctx.iface = init_interface(argv[1], SCPROTO_DSLCONF)) == NULL) {
		fprintf(stderr, "initialisation failed\n");
		return 1;
	}
	sem_init(&ctx.datalock, 0, 1);
	dslconf_print_init(ctx.iface, loglevel);

	if (pthread_create(&notifier, NULL, notifier_thread, &ctx))
		return 1;

	if (pthread_create(&dslconfd, NULL, dslconfd_thread, &ctx))
		return 1;

	while (1) {
		if (pthread_kill(notifier, 0) || pthread_kill(dslconfd, 0))
			return 1;
		sleep(1);
	}
	/* not reached */
	exit_interface(ctx.iface);
	return 0;
}
