#define _BSD_SOURCE
#define _POSIX_SOURCE
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/route.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "simplecom.h"
#include "dslconf.h"
#include "pidof.h"
#include "server.h"
#include "stealip.h"

/** Handle a NOP request by sending an ACK. */
static int nop_update(void)
{
	static int first_run = 1;
	if (first_run) {
		first_run = 0;
		return 1;
	}
	return 0;
}

static void nop_notify(struct interface iface)
{
	send_ack(iface, REQ_NOP);
}

static int nop_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	return send_ack(iface, pkt->hdr.type);
}

DSLCONF_CMD(NOP, nop);

struct ipconfig {
	char iface[IFNAMSIZ];
	char addr[16];
	char mask[16];
};

/** Parse a setip compatible request command line.
 * @param[in] arg Commandline to parse.
 * @return Instance of struct ipconfig filled with the parsed data, or NULL on error.
 *
 * The commandline may be in one of the following formats:
 * 1) <iface name> <IP address in dot-notation>/<Prefix>
 * 2) <iface name> <IP address in dot-notation> <Netmask in dot-notation> */
static struct ipconfig *parse_setip_arg(const struct dslconf_pkt *pkt)
{
	unsigned int ip[4], mask[4], i;
	struct ipconfig *ipcp;

	ipcp = malloc(sizeof(struct ipconfig));
	if (!ipcp)
		return NULL;
	memset(ipcp, 0, sizeof(*ipcp));

	if (sscanf(pkt->data, "%s %u.%u.%u.%u/%u", ipcp->iface,
	           &ip[0], &ip[1], &ip[2], &ip[3], &mask[0]) == 6) {
		/* convert prefix length to bitmask */
		for (i = 0; i < mask[0]; i++)
			mask[1] = (mask[1] << 1) + 1;
		for (i = mask[0]; i < 32; i++)
			mask[1] <<= 1;
		/* and finally to subnet mask
		 * (bitmask in mask[1], so overwrite this at last) */
		mask[0] = (mask[1] >> 24) & 0xff;
		mask[3] = mask[1] & 0xff;
		mask[2] = (mask[1] >> 8) & 0xff;
		mask[1] = (mask[1] >> 16) & 0xff;
	} else if (sscanf(pkt->data, "%s %u.%u.%u.%u %u.%u.%u.%u", ipcp->iface,
	           &ip[0], &ip[1], &ip[2], &ip[3],
	           &mask[0], &mask[1], &mask[2], &mask[3]) == 9) {
		/* nothing to do here */
	} else {
		free(ipcp);
		return NULL;
	}

	snprintf(ipcp->addr, 16, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
	snprintf(ipcp->mask, 16, "%u.%u.%u.%u",
			mask[0], mask[1], mask[2], mask[3]);
	return ipcp;
}

/** Handle a SETIP request, i.e. set the IP of a given interface. */
static int setip_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	int sockfd;
	struct ifreq ifr;
	struct ipconfig *ipcp;
	struct sockaddr_in sin;
	const char *err_msg = NULL;

#define FAIL(label, msg) { err_msg = msg; goto fail_##label; }

	if (!(ipcp = parse_setip_arg(pkt)))
		return send_nack_msg(iface, pkt->hdr.type, "illegal parameters");

	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
		FAIL(free_ipcp, strerror(errno));

	dnotify("setting IP of %s to %s mask %s", ipcp->iface, ipcp->addr, ipcp->mask);

	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ipcp->iface);
	memset(&sin, 0, sizeof(struct sockaddr));
	sin.sin_family = AF_INET;
	if (!inet_aton(ipcp->addr, &sin.sin_addr))
		FAIL(close_sockfd, "given IP not accepted");
	memcpy(&ifr.ifr_addr, &sin, sizeof(struct sockaddr));

	if (ioctl(sockfd, SIOCSIFADDR, &ifr) == -1)
		FAIL(close_sockfd, strerror(errno));

	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, IFNAMSIZ, "%s", ipcp->iface);
	memset(&sin, 0, sizeof(struct sockaddr));
	sin.sin_family = AF_INET;
	if (!inet_aton(ipcp->mask, &sin.sin_addr))
		FAIL(close_sockfd, "given netmask not accepted");
	memcpy(&ifr.ifr_netmask, &sin, sizeof(struct sockaddr));

	if (ioctl(sockfd, SIOCSIFNETMASK, &ifr) == -1)
		FAIL(close_sockfd, strerror(errno));

	close(sockfd);
	free(ipcp);
	return send_ack(iface, pkt->hdr.type);

fail_close_sockfd:
	close(sockfd);
fail_free_ipcp:
	free(ipcp);
	return send_nack_msg(iface, pkt->hdr.type, err_msg);
#undef FAIL
}
SIMPLE_DSLCONF_CMD(SET_IP, setip);

static char gw_iface[64], gw_ip[64];
static int gw_count;

/* Iface Destination Gateway Flags RefCnt Use Metric Mask MTU Window IRTT */
#define RT_ENTRY_FORMAT "%63s %lx %lx %x %*d %*d %*d %lx %*d %*d %*d\n"

/* code taken from the conky project (a snippet i've written on my own) */
static int get_gw_info(void)
{
	FILE *fp;
	struct in_addr ina;
	char iface[64];
	unsigned long dest, gate, mask;
	unsigned int flags;

	memset(gw_iface, 0, 64);
	memset(gw_ip, 0, 64);
	gw_count = 0;

	if (!(fp = fopen("/proc/net/route", "r"))) {
		return 1;
	}

	/* skip over the table header line, which is always present */
	fscanf(fp, "%*[^\n]\n");

	while (!feof(fp)) {
		if(fscanf(fp, RT_ENTRY_FORMAT,
			  iface, &dest, &gate, &flags, &mask) != 5) {
			puts("failure parsing /proc/net/route!");
			break;
		}
		if (!(dest || mask) && ((flags & RTF_GATEWAY) || !gate) ) {
			gw_count++;
			snprintf(gw_iface, 64, "%s", iface);
			ina.s_addr = gate;
			snprintf(gw_ip, 64, "%s", inet_ntoa(ina));
		}
	}
	fclose(fp);
	return 0;
}

/** Return the IP of an interface. */
static int getip_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	struct ifreq ifr;
	int sockfd;

	sockfd = socket(AF_INET, SOCK_DGRAM, 0);

	if (!strcmp(pkt->data, "gwiface")) {
		get_gw_info();
		if (!gw_count)
			return send_nack_msg(iface, pkt->hdr.type,
					"no default route present");
		snprintf(ifr.ifr_name, IFNAMSIZ, "%s", gw_iface);
	} else
		snprintf(ifr.ifr_name, IFNAMSIZ, "%s", pkt->data);

	ioctl(sockfd, SIOCGIFADDR, &ifr);
	close(sockfd);

	return send_ack_msg(iface, pkt->hdr.type,
			inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
}
SIMPLE_DSLCONF_CMD(GET_IP, getip);

/** Return either the gateway's IP address
 * or the interface name the gateway is reachable via. */
static int gw_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	if (get_gw_info())
		return send_nack_msg(iface, pkt->hdr.type, "get_gw_info() failed!");

	if (pkt->data[0] == 'i' && pkt->data[1] == 'f')
		return send_ack_msg(iface, pkt->hdr.type, gw_iface);
	if (pkt->data[0] == 'i' && pkt->data[1] == 'p')
		return send_ack_msg(iface, pkt->hdr.type, gw_ip);
	if (pkt->data[0] == 'c' && pkt->data[1] == 'o') {
		char buf[23];
		snprintf(buf, 23, "%d", gw_count);
		return send_ack_msg(iface, pkt->hdr.type, buf);
	}
	return send_nack_msg(iface, pkt->hdr.type, "unknown argument passed");
}
SIMPLE_DSLCONF_CMD(GET_GW, gw);

/** Try to kill and respawn a given commandline. */
static int reload_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char *p = NULL;
	int pid;

	if ((p = index(pkt->data, ' ')))
		(*p) = '\0';
	pid = pidof(pkt->data);
	if (p)
		(*p) = ' ';

	if (pid <= 0)
		return send_nack_msg(iface, pkt->hdr.type,
		                     "no running command with that name found");
	if (kill(pid, SIGTERM))
		return send_nack_msg(iface, pkt->hdr.type,
		                     "failed to kill() running process");

	if (system(pkt->data))
		return send_nack_msg(iface, pkt->hdr.type, "system() or command failed");

	return send_ack(iface, pkt->hdr.type);
}
SIMPLE_DSLCONF_CMD(REEXEC, reload);

/* Read contents of file pointed to by fp into newly allocated buffer,
 * returning a pointer to it. Caller has to free returned memory. */
char *read_fp_to_buf(FILE *fp)
{
	char *buf = NULL;
	int buflen = 0, offset = 0, rlen;

	while (fp && !feof(fp)) {
		if (offset >= buflen - 1) {
			buflen += 8192;
			buf = realloc(buf, buflen);
			if (!buf)
				return buf;
		}
		rlen = fread(buf + offset, 1, buflen - offset - 1, fp);
		offset += rlen;	/* offset will get (buflen - 1) at max */
		buf[offset] = 0;
	}

	return buf;
}

char *get_cmd_output(const char *cmd)
{
	FILE *fp;
	char *buf = NULL;

	if ((fp = popen(cmd, "r"))) {
		buf = read_fp_to_buf(fp);
		pclose(fp);
	}
	return buf;
}

/** Call some command and return it's output as ACK message. */
static int cmd_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char *buf;

	dnotify("command callback being called, executing '%s'", pkt->data);

	if (!(buf = get_cmd_output(pkt->data))) {
		send_nack_errno(iface, pkt->hdr.type);
		return 0;
	}
	send_ack_msg(iface, pkt->hdr.type, buf);

	free(buf);
	return 0;
}
SIMPLE_DSLCONF_CMD(EXEC, cmd);

/** Return the value of the given environment variable name. */
static int getenv_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char *val;

	dnotify("calling getenv for '%s'", pkt->data);
	if ((val = getenv(pkt->data)) == NULL)
		return send_nack(iface, pkt->hdr.type);

	dnotify("returning value '%s' for val '%s'", val, pkt->data);
	return send_ack_msg(iface, pkt->hdr.type, val);
}
SIMPLE_DSLCONF_CMD(GET_ENV, getenv);

/** Define a given environment variable to the given value. */
static int setenv_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char *val;

	if ((val = index(pkt->data, '=')) == NULL)
		return send_nack_msg(iface, pkt->hdr.type,
				"invalid syntax (no equal sign found)");

	*val = 0;
	if (setenv(pkt->data, val + 1, 1))
		return send_nack_msg(iface, pkt->hdr.type, "setenv() failed");

	return send_ack(iface, pkt->hdr.type);
}
SIMPLE_DSLCONF_CMD(SET_ENV, setenv);

/* two IPs of max 16 chars, one whitespace, one slash, a max 2 digits prefix len and a trailing \0 */
#define IPMAX (16 * 2) + 1 + 1 + 2 + 1
static char pub_ip[IPMAX];

static int get_pub_ip_update(void)
{
	char ip[IPMAX];
	int fd, iplen;

	if (access(PUB_IP_FILE, F_OK))
		/* treat missing file as OK */
		return 0;

	if ((fd = open(PUB_IP_FILE, O_RDONLY)) == -1) {
		derror("could not open " PUB_IP_FILE);
		return -1;
	}

	if ((iplen = read(fd, ip, IPMAX - 1)) < 0) {
		derror("read() failed: %s", strerror(errno));
		iplen = 0;
	}
	ip[iplen] = '\0';
	close(fd);

	if (strncmp(ip, "none", 4)) {
		while (iplen > 0 && !isdigit(ip[iplen - 1]))
			ip[--iplen] = 0;
	}

	if (strcmp(pub_ip, ip)) {
		strncpy(pub_ip, ip, IPMAX);
		dnotify("saving new public IP: '%s'", pub_ip);
		return 1;
	}
	return 0;
}

static void get_pub_ip_notify(struct interface iface)
{
	send_ack_msg(iface, REQ_GET_PUB_IP, pub_ip);
}

static int get_pub_ip_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	if (!strlen(pub_ip) || !strncmp(pub_ip, "none", 4))
		return send_nack_msg(iface, pkt->hdr.type, "not available");

	return send_ack_msg(iface, pkt->hdr.type, pub_ip);
}

DSLCONF_CMD(GET_PUB_IP, get_pub_ip);

static int version_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	return send_ack_msg(iface, pkt->hdr.type, VERSION);
}
SIMPLE_DSLCONF_CMD(GET_VERSION, version);

char *fsstamp = NULL;
char *scmver = NULL;

static int read_fsstamp_file(void)
{
	FILE *fp;
	char buf[128];

	if ((fp = fopen(FSSTAMP_FILE, "r")) == NULL)
		return 1;

	while (fgets(buf, 128, fp)) {
		if (!strncmp(buf, "FSSTAMP=", 8)) {
			if (fsstamp)
				free(fsstamp);
			fsstamp = calloc(strlen(buf + 8), 1);
			strncpy(fsstamp, buf + 8, strlen(buf + 8) - 1);
		} else if (!strncmp(buf, "SCMVER=", 7)) {
			if (scmver)
				free(scmver);
			scmver = calloc(strlen(buf + 7), 1);
			strncpy(scmver, buf + 7, strlen(buf + 7) - 1);
		}
	}
	fclose(fp);
	return 0;
}

static int fs_stamp_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	if (!fsstamp && read_fsstamp_file())
		return send_nack_msg(iface, pkt->hdr.type,
				"failed to open " FSSTAMP_FILE);
	if (!fsstamp)
		return send_nack_msg(iface, pkt->hdr.type,
				"no timestamp found in " FSSTAMP_FILE);
	return send_ack_msg(iface, pkt->hdr.type, fsstamp);
}
SIMPLE_DSLCONF_CMD(FS_STAMP, fs_stamp);

static int scmver_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	if (!scmver && read_fsstamp_file())
		return send_nack_msg(iface, pkt->hdr.type,
				"failed to open " FSSTAMP_FILE);
	if (!scmver)
		return send_nack_msg(iface, pkt->hdr.type,
				"no SCM version found in " FSSTAMP_FILE);
	return send_ack_msg(iface, pkt->hdr.type, scmver);
}
SIMPLE_DSLCONF_CMD(SCMVER, scmver);

static int get_pppd_log_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	FILE *fp;
	char *buf;

	dnotify("pppd_log callback being called");

	if ((fp = fopen(PPPD_LOGFILE, "r")) == NULL)
		return send_nack_msg(iface, pkt->hdr.type, "failed to open " PPPD_LOGFILE);

	buf = read_fp_to_buf(fp);
	send_ack_msg(iface, pkt->hdr.type, buf);

	if (buf)
		free(buf);
	fclose(fp);
	return 0;
}
SIMPLE_DSLCONF_CMD(GET_PPPD_LOG, get_pppd_log);

